# SPDX-FileCopyrightText: 2024 Deutsche Klimarechenzentrum
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module that handles S3 connections using Minio with parallel uploads."""

import asyncio
import logging
import os
from pathlib import Path
from typing import List, Optional, Union

import aioboto3
from aiobotocore.session import ClientCreatorContext
from boto3.s3.transfer import TransferConfig
from botocore.exceptions import (
    NoCredentialsError,
    PartialCredentialsError,
    ProfileNotFound,
)

logging.getLogger("boto3").setLevel(logging.WARNING)
logging.getLogger("botocore").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

__all__ = ["S3"]


class S3:
    """Establish a connection to an S3-compatible storage (e.g., Minio).
        Push files or directories to a speficied bucket

        Parameters
        ----------
        bucket_name: str
            The name of the bucket to interact with.
        profile_name: str
            The name of the profile to use from AWS_SHARED_CONFIG.
            It should contain the access_key and secret_key, region
            sand have an endpoint_url defined. See example bellow
        endpoint_url: str
            The endpoint URL of the S3-compatible storage.
        access_key: str
            Access key for authentication.
        secret_key: str
            Secret key for authentication.

        Examples
        --------
        Content of AWS_SHARED_CONFIG:
        ```
    [profile myprofile-name]
    aws_access_key_id = myaccesskeuy
    aws_secret_access_key = mysecret
    region = eu-dkrz-1
    endpoint_url = https://s3.eu-dkrz-1.dkrz.cloud
        ```
    """

    def __init__(
        self,
        bucket_name: str,
        profile_name: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
    ):
        """Initialize the S3Handler."""
        self.endpoint_url = endpoint_url or os.environ.get("AWS_ENDPOINT_URL")
        self.access_key = access_key or os.environ.get("AWS_ACCESS_KEY_ID")
        self.secret_key = secret_key or os.environ.get("AWS_SECRET_ACCESS_KEY")
        self.region = os.environ.get("AWS_REGION") or os.environ.get(
            "AWS_DEFAULT_REGION", "eu-dkrz-1"
        )
        self.profile = profile_name or os.environ.get("AWS_PROFILE")
        self.bucket_name = bucket_name
        self.semaphore: asyncio.Semaphore
        try:
            if profile_name is not None:
                try:
                    self.session = aioboto3.Session(profile_name=self.profile)
                    logger.debug("Using user defined S3 profile/credential")
                except ProfileNotFound:
                    os.environ["AWS_CONFIG_FILE"] = (
                        os.environ["EVALUATION_SYSTEM_CONFIG_DIR"]
                        + "/s3/config"
                    )
                    self.session = aioboto3.Session(profile_name=self.profile)
            else:
                self.session = aioboto3.Session(
                    aws_access_key_id=self.access_key,
                    aws_secret_access_key=self.secret_key,
                    region_name=self.region,
                )

            self.client: ClientCreatorContext = None

        except (NoCredentialsError, PartialCredentialsError) as e:
            logger.error(f"Credential error: {e}")
            raise
        except Exception as e:
            logger.error(f"Failed to connect to S3: {e}")
            raise

        self.transfer_config = TransferConfig(
            multipart_threshold=16 * 1024 * 1024,
            multipart_chunksize=16 * 1024 * 1024,
            max_concurrency=10,
        )

    async def __aenter__(self):
        """
        Enter the context and create the persistent client.
        """
        if self.client is None:
            if self.endpoint_url is None:
                self.client = await self.session.client("s3").__aenter__()
            else:
                self.client = await self.session.client(
                    "s3", endpoint_url=self.endpoint_url
                ).__aenter__()
            self.semaphore = asyncio.Semaphore(100)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the context and close the persistent client.
        """
        if self.client is not None:
            await self.client.__aexit__(exc_type, exc_val, exc_tb)
            self.client = None

    async def _run_context_manager(self, coro, *args, **kwargs):
        """
        Utility to run async operations in a context manager.
        """
        async with self:
            return await coro(*args, **kwargs)

    async def upload_file_async(
        self, file_path: Union[str, Path], key: str
    ) -> str:
        async with self.semaphore:
            try:
                await self.client.upload_file(
                    str(file_path), self.bucket_name, Key=key
                )
                logger.info(f"Successfully uploaded {file_path} to {key}.")
            except Exception as e:
                logger.error(f"Error uploading {file_path} to {key}: {e}")
        return await self.get_file_url_async(key)

    def upload_file(self, file_path: Union[str, Path], key: str) -> str:
        return asyncio.run(
            self._run_context_manager(self.upload_file_async, file_path, key)
        )

    async def upload_directory_async(
        self,
        directory: Path,
        prefix: Union[str, Path] | None = None,
    ) -> List[str]:
        """
        Async directory upload.

        :param directory: Path to the local source directory
        :param prefix: Optional prefix for the objects
        :returns: List containing links for each object, last link for listing the content of bucket
        """
        tasks = []
        for root, _, files in os.walk(directory):
            for file_name in files:
                file_path = Path(root) / file_name
                relative_path = file_path.relative_to(directory.parent)
                key = str(Path(prefix or "") / relative_path)
                tasks.append(self.upload_file_async(file_path, key))
        links = await asyncio.gather(*tasks)
        # generate presigned url for directory listing
        listing_url = await self.client.generate_presigned_url(
            "list_objects",
            Params={
                "Bucket": self.bucket_name,
                "Prefix": (
                    str(Path(prefix) / directory.name) + "/" if prefix else ""
                ),
            },
            ExpiresIn=604800,
        )

        links.append(listing_url)
        return links

    def upload_directory(
        self,
        directory: Path,
        prefix: Union[str, Path] | None = None,
    ) -> List[str]:
        """
        Uploads the contents of a local directory to an S3 bucket synchronously.

        :param directory_path: Path to the local directory to upload.
        :param prefix: Optional prefix to prepend to the uploaded objects' keys.
        :returns: List containing links for each object, last link for listing the content of bucket
        """
        return asyncio.run(
            self._run_context_manager(
                self.upload_directory_async, Path(directory), prefix
            )
        )

    def upload(
        self, inp_path: Path, prefix: Union[str, Path] | None = None
    ) -> Optional[str]:
        if inp_path.is_file():
            key = str(Path(prefix or "") / inp_path.name)
            url = self.upload_file(inp_path, key)
            print(f"Uploaded file: {url}")
        elif inp_path.is_dir():
            urls: List = self.upload_directory(inp_path, prefix=prefix)
            url = urls.pop()
            print(f"Uploaded files:\n {urls}")
            print(f"List the contents of the bucket: {url}")
        else:
            print(f"Invalid input path: {inp_path}")
            return None
        return url

    async def get_file_url_async(self, key: str) -> str:
        if key.startswith("public"):
            return (
                self.client.meta.endpoint_url.strip("/")
                + f"/{self.bucket_name}/{key.strip('/')}"
            )
        return await self.client.generate_presigned_url(
            "get_object",
            Params={"Bucket": self.bucket_name, "Key": key},
            ExpiresIn=604800,
        )

    def get_file_url(self, key: str) -> str:
        return asyncio.run(
            self._run_context_manager(self.get_file_url_async, key)
        )


if __name__ == "__main__":
    import sys

    handler = S3(profile_name="nukleus", bucket_name="ch1187")

    # Example usage: Upload a file or directory
    inp_path = Path(sys.argv[1]).expanduser().absolute()
    handler.upload(inp_path, prefix="public/")
